"""The main Module class."""
import json
import typing as t

import lovely_tensors as lt  # type: ignore
import torch
import yaml
from rich.console import Console
from rich.panel import Panel
from rich.syntax import Syntax
from rich.table import Table

from torchsilk import type_defs as tdf
from torchsilk.modules.buffers import NamedBuffers
from torchsilk.modules.module_state import ModuleState
from torchsilk.modules.type_defs import FunctionalModule, FunctionalModuleWithoutBuffers


def json_to_yaml(x: str) -> str:
    """Convert a json string to yaml."""
    return yaml.safe_dump(json.loads(x))


class ModuleSetupNotImplemented(NotImplementedError):
    """Raised when Module.__setup_... functions are not implemented."""


class Module(t.Generic[tdf.PydanticModel, tdf.P, tdf.R_co]):
    """Opinionated wrapper around torch.nn.Module that acts like a stateless model.

    To construct a stateless model, you must implement either subclass either
    __setup_module_state__ or __setup_torch_module__. If you implement both, then
    __setup_module_state__ will be used.

    Examples:

    1. Model that doesn't have any parametrizable layers.

    >>> import torch
    ... import functorch as ft
    ... import torchsilk as tsk
    ... class MyModule(Module[tdf.EmptyModel, torch.Tensor, torch.Tensor]):
    ...     @classmethod
    ...     def __setup_torch_module__(cls, config) -> torch.nn.Module:
    ...         return torch.nn.Linear(1, 1)
    ... tsk_module = MyModule()
    ... torch_module = MyModule.__setup_torch_module__(tsk.BaseConfig())
    ... func_module, _ = ft.make_functional(torch_module)
    ... x = torch.randn(4, 1)
    ... assert torch.allclose(func_module(tsk_module.parameters(), x), tsk_module(x))

    2. Model that has parametrizable layers.

    >>> import torch
    ... import torchsilk as tsk
    ... from pydantic import validator
    ... class MyFeedForwardConfig(tsk.BaseConfig):
    ...     in_features: int
    ...     out_features: int
    ...     hidden_features: list[int]
    ...     hidden_activations: list[str | torch.nn.Module]
    ...     @validator('hidden_activations')
    ...     def check_activations(cls, v: t.Any) -> list[torch.nn.Module]:
    ...         return [getattr(torch.nn, a) for a in v]
    ... class MyFeedForward(Module[MyFeedForwardConfig, torch.Tensor, torch.Tensor]):
    ...     @classmethod
    ...     def __setup_torch_module__(self, config) -> torch.nn.Module:
    ...         layers = []
    ...         for i, (in_features, out_features, activation) in enumerate(
    ...             zip(
    ...                 [config.in_features, *config.hidden_features],
    ...                 [*config.hidden_features, config.out_features],
    ...                 [*config.hidden_activations, torch.nn.Identity],
    ...             )
    ...         ):
    ...             layers.append(torch.nn.Linear(in_features, out_features))
    ...             if activation != "linear":
    ...                 layers.append(activation())
    ...         return torch.nn.Sequential(*layers)
    ... config = MyFeedForwardConfig(
    ...     in_features=1,
    ...     out_features=1,
    ...     hidden_features=[2, 3],
    ...     hidden_activations=["ReLU", "ReLU"]
    ... )
    ... tsk_module = MyFeedForward(config)
    ... torch_module = MyFeedForward.__setup_torch_module__(config)
    ... func_module, _ = ft.make_functional(torch_module)
    ... x = torch.randn(4, config.in_features)
    ... assert torch.allclose(func_module(tsk_module.parameters(), x), tsk_module(x))

    3. Model that has buffers.

    >>> import torch
    ... import torchsilk as tsk
    ... class MyModule(Module[tdf.EmptyModel, [torch.Tensor], torch.Tensor]):
    ...     @classmethod
    ...     def __setup_torch_module__(cls, config) -> torch.nn.Module:
    ...         class _MyModule(torch.nn.Module):
    ...             def __init__(self):
    ...                 super().__init__()
    ...                 self.register_buffer("bias", torch.ones(1))
    ...             def forward(self, x):
    ...                 return x + self.bias
    ...         return _MyModule()
    ... tsk_module = MyModule(tdf.EmptyModel())
    ... torch_module = MyModule.__setup_torch_module__(tdf.EmptyModel())
    ... func_module, *_ = ft.make_functional_with_buffers(torch_module)
    ... x = torch.randn(4, 1)
    ... assert torch.allclose(
    ...     func_module(tsk_module.parameters(), tsk_module.buffers(), x),
    ...     tsk_module(x)
    ... )
    ... assert torch.allclose(tsk_module.buffers()[0], torch_module.bias)
    ... assert torch.allclose(tsk_module.named_buffers().bias, torch_module.bias)
    """

    def __init__(
        self,
        config: tdf.BaseConfig = tdf.BaseConfig(),
        module_state: ModuleState[tdf.P, tdf.R_co] | None = None,
    ) -> None:
        self.config = config
        if module_state:
            self._module_state = module_state
        else:
            self._init_module_state()

    def _init_module_state(self) -> None:
        try:
            self._module_state = self.__setup_module_state__(self.config)
        except ModuleSetupNotImplemented:
            try:
                torch_module = self.__setup_torch_module__(self.config)
            except ModuleSetupNotImplemented:
                raise ModuleSetupNotImplemented(
                    "Both __setup_module_state__ and __setup_torch_module__ are not"
                    + " implemented."
                ) from None
            else:
                self._module_state = ModuleState[tdf.P, tdf.R_co].from_torch_module(
                    torch_module
                )

    @classmethod
    def __setup_module_state__(
        cls, config: tdf.PydanticModel
    ) -> ModuleState[tdf.P, tdf.R_co]:
        raise ModuleSetupNotImplemented(
            f"{cls}.__setup_module_state__ not implemented."
        )

    @classmethod
    def __setup_torch_module__(cls, config: tdf.PydanticModel) -> torch.nn.Module:
        raise ModuleSetupNotImplemented(
            f"{cls}.__setup_torch_module__ not implemented."
        )

    @property
    def module(
        self,
    ) -> FunctionalModuleWithoutBuffers[tdf.P, tdf.R_co] | FunctionalModule[
        tdf.P, tdf.R_co
    ]:
        return self._module_state.module

    def apply(
        self,
        params: tuple[torch.nn.Parameter, ...],
        *args: tdf.P.args,
        **kwargs: tdf.P.kwargs,
    ) -> tdf.R_co:
        """Apply the module to the given arguments.

        Args:
        -----
            params: Parameters of the module.
            buffers: Buffers of the module.
            *args: Arguments to the module.
            **kwargs: Keyword arguments to the module.
        Returns:
        --------
            The result of applying the module to the given arguments.

        Examples:

        1. Module without buffers.
        >>> import torch
        ... import functorch as ft
        ... import torchsilk as tsk
        ... class MyModule(tsk.Module[tdf.EmptyModel, [torch.Tensor], torch.Tensor]):
        ...     @classmethod
        ...     def __setup_torch_module__(cls, config) -> torch.nn.Module:
        ...         return torch.nn.Linear(1, 1)
        ... tsk_module = MyModule(tdf.EmptyModel())
        ... torch_module = MyModule.__setup_torch_module__(tdf.EmptyModel())
        ... func_module, *_ = ft.make_functional(torch_module)
        ... x = torch.randn(4, 1)
        ... assert torch.allclose(
        ...     func_module(tsk_module.parameters(), x),
        ...     tsk_module(x)
        ... )
        ... new_params = tuple(torch.nn.Parameter(torch.randn(1, 1)) for _ in range(2))
        ... assert torch.allclose(
        ...     func_module(new_params, x),
        ...     tsk_module.apply(new_params, x)
        ... )

        2. Module with buffers.
        >>> import torch
        ... import functorch as ft
        ... import torchsilk as tsk
        ... class MyModule(tsk.Module[tdf.EmptyModel, [torch.Tensor], torch.Tensor]):
        ...     @classmethod
        ...     def __setup_torch_module__(cls, config) -> torch.nn.Module:
        ...         class _MyModule(torch.nn.Module):
        ...             def __init__(self):
        ...                 super().__init__()
        ...                 self.linear = torch.nn.Linear(1, 1)
        ...                 self.register_buffer("bias", torch.ones(1))
        ...             def forward(self, x):
        ...                 return self.linear(x) + self.bias
        ...         return _MyModule()
        ... tsk_module = MyModule(tdf.EmptyModel())
        ... torch_module = MyModule.__setup_torch_module__(tdf.EmptyModel())
        ... func_module, *_ = ft.make_functional_with_buffers(torch_module)
        ... x = torch.randn(4, 1)
        ... assert torch.allclose(
        ...     func_module(tsk_module.parameters(), tsk_module.buffers(), x),
        ...     tsk_module(x)
        ... )
        ... new_params = tuple(torch.nn.Parameter(torch.randn(1, 1)) for _ in range(2))
        ... assert torch.allclose(
        ...     func_module(new_params, tsk_module.buffers(), x),
        ...     tsk_module.apply(new_params, x)
        ... )
        """
        return self._module_state.apply(params, *args, **kwargs)

    def apply_with_buffers(
        self,
        params: tuple[torch.nn.Parameter, ...],
        buffers: tuple[torch.Tensor, ...],
        *args: tdf.P.args,
        **kwargs: tdf.P.kwargs,
    ) -> tdf.R_co:
        """Apply the module to the given arguments.

        Args:
        -----
            params: Parameters of the module.
            buffers: Buffers of the module.
            *args: Arguments to the module.
            **kwargs: Keyword arguments to the module.
        Returns:
        --------
            The result of applying the module to the given arguments.

        Examples:
        ---------

        1. Module without buffers.
        >>> import torch
        ... import functorch as ft
        ... import pytest
        ... import torchsilk as tsk
        ... class MyModule(tsk.Module[tdf.EmptyModel, [torch.Tensor], torch.Tensor]):
        ...     @classmethod
        ...     def __setup_torch_module__(cls, config) -> torch.nn.Module:
        ...         return torch.nn.Linear(1, 1)
        ... tsk_module = MyModule(tdf.EmptyModel())
        ... torch_module = MyModule.__setup_torch_module__(tdf.EmptyModel())
        ... func_module, *_ = ft.make_functional(torch_module)
        ... x = torch.randn(4, 1)
        ... buffers = tuple()
        ... with pytest.raises(ValueError):
        ...     tsk_module.apply_with_buffers(tsk_module.parameters(), buffers, x)

        2. Module with buffers.
        >>> import torch
        ... import functorch as ft
        ... import torchsilk as tsk
        ... class MyModule(tsk.Module[tdf.EmptyModel, [torch.Tensor], torch.Tensor]):
        ...     @classmethod
        ...     def __setup_torch_module__(cls, config) -> torch.nn.Module:
        ...         class _MyModule(torch.nn.Module):
        ...             def __init__(self):
        ...                 super().__init__()
        ...                 self.linear = torch.nn.Linear(1, 1)
        ...                 self.register_buffer("bias", torch.ones(1))
        ...             def forward(self, x):
        ...                 return self.linear(x) + self.bias
        ...         return _MyModule()
        ... tsk_module = MyModule(tdf.EmptyModel())
        ... torch_module = MyModule.__setup_torch_module__(tdf.EmptyModel())
        ... func_module, *_ = ft.make_functional_with_buffers(torch_module)
        ... x = torch.randn(4, 1)
        ... assert torch.allclose(
        ...     func_module(tsk_module.parameters(), tsk_module.buffers(), x),
        ...     tsk_module(x)
        ... )
        ... new_params = tuple(torch.nn.Parameter(torch.randn(1, 1)) for _ in range(2))
        ... assert torch.allclose(
        ...     func_module(new_params, tsk_module.buffers(), x),
        ...     tsk_module.apply(new_params, x)
        ... )
        """
        return self._module_state.apply_with_buffers(params, buffers, *args, **kwargs)

    def parameters(self) -> tuple[torch.nn.Parameter, ...]:
        return self._module_state.parameters

    def buffers(self) -> tuple[torch.Tensor, ...]:
        return self._module_state.buffers

    def named_buffers(self) -> NamedBuffers:
        return self._module_state.named_buffers

    def to(
        self, *args: t.Any, **kwargs: t.Any
    ) -> "Module[tdf.PydanticModel, tdf.P, tdf.R_co]":
        """Move the module to the specified device, dtype and/or memory format.

        Args:
        -----
            *args: t.Any
            **kwargs: t.Any
            Any arguments accepted by torch.Tensor.to(...).

        Returns:
        --------
            Module[tdf.PydanticModel, tdf.P, tdf.R]
            A copy of the module with the specified device, dtype and/or memory format.

        Examples:
        ---------

        >>> import torch
        ... import functorch as ft
        ... import torchsilk as tsk
        ... class MyModule(tsk.Module[tdf.EmptyModel, [torch.Tensor], torch.Tensor]):
        ...     @classmethod
        ...     def __setup_torch_module__(cls, config) -> torch.nn.Module:
        ...         return torch.nn.Linear(1, 1)
        ... tsk_module = MyModule(tdf.EmptyModel())
        ... torch_module = MyModule.__setup_torch_module__(tdf.EmptyModel())
        ... float64_module = tsk_module.to(torch.float64)
        ... float64_torch_module = torch_module.to(torch.float64)
        ... func_module, _ = ft.make_functional(float64_torch_module)
        ... x = torch.randn(4, 1).to(torch.float64)
        ... assert torch.allclose(
        ...     float64_module(x), func_module(float64_module.parameters(), x)
        ... )
        """
        return self.__class__(
            config=self.config, module_state=self._module_state.to(*args, **kwargs)
        )

    def evolve(
        self,
        parameters: tuple[torch.nn.Parameter, ...] | None = None,
        buffers: tuple[torch.Tensor, ...] | None = None,
        named_buffers: NamedBuffers | None = None,
    ) -> "Module[tdf.PydanticModel, tdf.P, tdf.R_co]":
        """Return new Module with same config but with new module_state from updates.

        Args:
        -----
            parameters: tuple[torch.nn.Parameter, ...]
                New parameters.
            buffers: tuple[torch.Tensor, ...]
                New buffers. Cannot be provided together with named_buffers.
            named_buffers: NamedBuffers
                New named_buffers. Cannot be provided together with buffers.
        Returns:
        --------
            Module[tdf.PydanticModel, tdf.P, tdf.R]
                New Module with updated parameters and/or buffers and/or named_buffers.

        Examples:
        ---------

        1. Update parameters.
        >>> import torch
        ... import functorch as ft
        ... import torchsilk as tsk
        ... class MyModule(Module[tdf.EmptyModel, [torch.Tensor], torch.Tensor]):
        ...     @classmethod
        ...     def __setup_torch_module__(cls, config) -> torch.nn.Module:
        ...         class _MyModule(torch.nn.Module):
        ...             def __init__(self):
        ...                 super().__init__()
        ...                 self.linear = torch.nn.Linear(1, 1)
        ...             def forward(self, x):
        ...                 return self.linear(x)
        ...         return _MyModule()
        ... tsk_module = MyModule(tdf.EmptyModel())
        ... torch_module = MyModule.__setup_torch_module__(tdf.EmptyModel())
        ... func_module, *_ = ft.make_functional(torch_module)
        ... x = torch.randn(4, 1)
        ... assert torch.allclose(
        ...     func_module(tsk_module.parameters(), x),
        ...     tsk_module(x)
        ... )
        ... new_params = (torch.nn.Parameter(torch.zeros(1)),)
        ... new_tsk_module = tsk_module.evolve(parameters=new_params)
        ... assert not torch.allclose(
        ...    func_module(tsk_module.parameters(), x),
        ...    new_tsk_module(x)
        ... )

        2. Update buffers.
        >>> import torch
        ... import torchsilk as tsk
        ... class MyModule(Module[tdf.EmptyModel, [torch.Tensor], torch.Tensor]):
        ...     @classmethod
        ...     def __setup_torch_module__(cls, config) -> torch.nn.Module:
        ...         class _MyModule(torch.nn.Module):
        ...             def __init__(self):
        ...                 super().__init__()
        ...                 self.register_buffer("bias", torch.ones(1))
        ...             def forward(self, x):
        ...                 return x + self.bias
        ...         return _MyModule()
        ... tsk_module = MyModule(tdf.EmptyModel())
        ... torch_module = MyModule.__setup_torch_module__(tdf.EmptyModel())
        ... func_module, *_ = ft.make_functional_with_buffers(torch_module)
        ... x = torch.randn(4, 1)
        ... assert torch.allclose(
        ...     func_module(tsk_module.parameters(), tsk_module.buffers(), x),
        ...     tsk_module(x)
        ... )
        ... new_buffers = (torch.zeros(1),)
        ... new_tsk_module = tsk_module.evolve(buffers=new_buffers)
        ... assert not torch.allclose(
        ...    func_module(tsk_module.parameters(), tsk_module.buffers(), x),
        ...    new_tsk_module(x)
        ... )

        3. Updating both parameters and buffers raises a ValueError.
        >>> import torch
        ... import torchsilk as tsk
        ... import pytest
        ... class MyModule(Module[tdf.EmptyModel, [torch.Tensor], torch.Tensor]):
        ...     @classmethod
        ...     def __setup_torch_module__(cls, config) -> torch.nn.Module:
        ...         class _MyModule(torch.nn.Module):
        ...             def __init__(self):
        ...                 super().__init__()
        ...                 self.linear = torch.nn.Linear(1, 1)
        ...                 self.register_buffer("bias", torch.ones(1))
        ...             def forward(self, x):
        ...                 return self.linear(x) + self.bias
        ...         return _MyModule()
        ... tsk_module = MyModule(tdf.EmptyModel())
        ... new_params = (torch.nn.Parameter(torch.zeros(1)),)
        ... new_buffers = (torch.zeros(1),)
        ... new_named_buffers = tsk_module.named_buffers().evolve_from_values(
        ...     new_buffers
        ... )
        ... with pytest.raises(ValueError):
        ...     tsk_module.evolve(
        ...         parameters=new_params,
        ...         buffers=new_buffers,
        ...         named_buffers=new_named_buffers,
        ...     )
        """
        return self.__class__(
            config=self.config,
            module_state=self._module_state.evolve(
                parameters=parameters, buffers=buffers, named_buffers=named_buffers
            ),
        )

    def show(self, max_n_params: int = 2) -> None:
        """Show a pretty representation of self.

        Args:
        -----
            max_n_params: int
                Maximum number of parameters to show.

        Example representation:
        -----------------------

        >>> import torch
        ... class MyModule(Module[tdf.EmptyModel, [torch.Tensor], torch.Tensor]):
        ...     @classmethod
        ...     def __setup_torch_module__(cls, config) -> torch.nn.Module:
        ...         class _MyModule(torch.nn.Module):
        ...             def __init__(self):
        ...                 super().__init__()
        ...                 self.linear = torch.nn.Linear(2, 4)
        ...                 self.register_buffer("bias", torch.randn(5, 5))
        ...             def forward(self, x):
        ...                 return self.linear(x) + self.bias
        ...         return _MyModule()
        ... tsk_module = MyModule(tdf.EmptyModel())
        ... tsk_module.show()

        ──────────────────────── MyModule ────────────────────────
        ╭────────────────── Config: EmptyModel ──────────────────╮
        │ {}                                                     │
        │                                                        │
        ╰────────────────────────────────────────────────────────╯
        ╭──────────────────────── Module ────────────────────────╮
        │ FunctionalModuleWithBuffers(                           │
        │   (stateless_model): _MyModule(                        │
        │     (linear): Linear(in_features=2, out_features=4,    │
        │ bias=True)                                             │
        │   )                                                    │
        │ )                                                      │
        ╰────────────────────────────────────────────────────────╯
        ┏━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┓
        ┃                       Parameters                       ┃
        ┡━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┩
        │ Parameter[4, 2] n=8 x∈[-0.571, 0.663] μ=0.068 σ=0.412  │
        │  [[0.346, -0.571], [0.502, 0.071], [-0.143, -0.233],   │
        │                    [0.663, -0.090]]                    │
        │ Parameter[4] x∈[-0.681, 0.514] μ=0.081 σ=0.560 [0.490, │
        │                 0.514, -0.681, 0.001]                  │
        └────────────────────────────────────────────────────────┘
        ┏━━━━━━━━━━━━━━━━━━━━━━┳━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┓
        ┃ Name                 ┃             Buffers             ┃
        ┡━━━━━━━━━━━━━━━━━━━━━━╇━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┩
        │ bias                 │  tensor[5, 5] n=25 x∈[-3.056,   │
        │                      │     2.311] μ=-0.062 σ=1.044     │
        └──────────────────────┴─────────────────────────────────┘
        """
        console = Console()
        console.rule(f"{self.__class__.__name__}")
        console.print(
            Panel(
                Syntax(json_to_yaml(self.config.json()), lexer="yaml"),
                title=f"Config: {self.config.__class__.__name__}",
                border_style="blue",
            )
        )
        console.print(
            Panel(str(self._module_state.module), title="Module", border_style="yellow")
        )
        if max_n_params > 0:
            table = Table(show_header=True, header_style="bold magenta")
            table.add_column("Parameters", width=console.width, justify="center")
            for param in self._module_state.parameters[:max_n_params]:
                table.add_row(str(lt.lovely(param)))
            console.print(table)
        if self._module_state.has_buffers:
            console.print(
                self.named_buffers().as_table(
                    name_column_width=24, buffers_column_width=console.width - 24
                )
            )

    def __call__(self, *args: tdf.P.args, **kwargs: tdf.P.kwargs) -> tdf.R_co:
        return self._module_state(*args, **kwargs)



import abc
import typing as t
import warnings

import torch
from torch.distributions import constraints
from torch.distributions.utils import lazy_property

DistributionParamsT = t.TypeVar("DistributionParamsT", bound="DistributionParams")


class DistributionParams(abc.ABC):
    @property
    @abc.abstractmethod
    def batch_shape(self) -> torch.Size:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def event_shape(self) -> torch.Size:
        raise NotImplementedError

    @property
    def support(self) -> constraints.Constraint:
        raise NotImplementedError

    @abc.abstractmethod
    def astuple(self) -> tuple[torch.Tensor, ...]:
        raise NotImplementedError

    @property
    def arg_constraints(self) -> t.Dict[str, constraints.Constraint]:
        return {}

    def validate_sample(self, value: torch.Tensor) -> None:
        """Argument validation for methods such as `log_prob`, `cdf`, `icdf` etc.

        The rightmost dimensions of `value` must agree with `self.event_shape` and
        `self.batch_shape`.

        Raises:
            ValueError: If `value` is not a valid sample.
        """
        event_dim_start = len(value.size()) - len(self.event_shape)
        if value.size()[event_dim_start:] != self.event_shape:
            raise ValueError(
                "The right-most size of value must match event_shape:"
                + f" {value.size()} vs {self.event_shape}."
            )

        actual_shape = value.size()
        expected_shape = self.batch_shape + self.event_shape
        for i, j in zip(reversed(actual_shape), reversed(expected_shape)):
            if i != 1 and j != 1 and i != j:
                raise ValueError(
                    "Value is not broadcastable with batch_shape+event_shape:"
                    + f" {actual_shape} vs {expected_shape}."
                )

        try:
            support = self.support
        except NotImplementedError:
            warnings.warn(
                f"{self.__class__} does not define `support` to enable "
                + "sample validation. Please initialize the distribution with "
                + "`validate_args=False` to turn off validation."
            )
            return
        assert support is not None
        valid: torch.ByteTensor = support.check(value)
        if not valid.all():
            raise ValueError(
                "Expected value argument "
                f"({type(value).__name__} of shape {tuple(value.shape)}) "
                f"to be within the support ({repr(support)}) "
                f"of the distribution {repr(self)}, "
                f"but found invalid values:\n{value}"
            )

    def _validate_args(self) -> None:
        """Validate the parameters of the distribution.

        NOTE: Taken from torch.distributions.distribution.Distribution.__init__

        Extracted into a separate method on parameters because Distribution is stateless
        """
        if not self.arg_constraints:
            warnings.warn(
                f"{self.__class__} does not define `arg_constraints`. Not validating."
            )
        for param, constraint in self.arg_constraints.items():
            if constraints.is_dependent(constraint):
                continue  # skip constraints that cannot be checked
            if param not in self.__dict__ and isinstance(
                getattr(type(self), param), lazy_property
            ):
                continue  # skip checking lazily-constructed args
            value = getattr(self, param)
            valid: torch.ByteTensor = constraint.check(value)
            if not valid.all():
                raise ValueError(
                    f"Expected parameter {param} "
                    f"({type(value).__name__} of shape {tuple(value.shape)}) "
                    f"of distribution {repr(self)} "
                    f"to satisfy the constraint {repr(constraint)}, "
                    f"but found invalid values:\n{value}"
                )


class Distribution(abc.ABC, t.Generic[DistributionParamsT]):
    @classmethod
    @abc.abstractmethod
    def sample(
        cls,
        params: DistributionParamsT,
        sample_shape: torch.Size | tuple[int, ...] = torch.Size(),
        generator: torch.Generator | None = None,
    ) -> torch.Tensor:
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def rsample(
        cls,
        params: DistributionParamsT,
        sample_shape: torch.Size | tuple[int, ...] = torch.Size(),
        generator: torch.Generator | None = None,
    ) -> torch.Tensor:
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def log_prob(cls, params: DistributionParamsT, value: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def entropy(cls, params: DistributionParamsT) -> torch.Tensor:
        raise NotImplementedError

    @classmethod
    def perplexity(cls, params: DistributionParamsT) -> torch.Tensor:
        return torch.exp(cls.entropy(params))

    @classmethod
    def _extended_shape(
        cls,
        params: DistributionParamsT,
        sample_shape: torch.Size | tuple[int, ...] = torch.Size(),
    ) -> torch.Size:
        if not isinstance(sample_shape, torch.Size):
            sample_shape = torch.Size(sample_shape)
        result = sample_shape + params.batch_shape + params.event_shape
        if not isinstance(result, torch.Size):
            result = torch.Size(result)
        return result



import math

import attrs
import torch
from torch.distributions import constraints
from torch.distributions.utils import (
    broadcast_all,  # pyright: ignore[reportUnknownVariableType]
)

from torchsilk.distributions.base import Distribution, DistributionParams
from torchsilk.distributions.utils import (
    _standard_normal,  # pyright: ignore[reportPrivateUsage]
)


@attrs.define(frozen=True)
class NormalDistParams(DistributionParams):
    loc: torch.Tensor
    scale: torch.Tensor

    validate_args: bool = False

    _arg_constraints: dict[str, constraints.Constraint] = attrs.field(
        init=False,
        factory=lambda: {
            "loc": constraints.real,
            "scale": constraints.positive,
        },
    )
    _batch_shape: torch.Size = attrs.field(init=False)
    _event_shape: torch.Size = attrs.field(init=False)

    def __attrs_post_init__(self):
        if self.loc.shape != self.scale.shape:
            raise ValueError(
                "loc and scale must have the same shape."
                + f" Got {self.loc.shape} and {self.scale.shape}."
            )
        if self.validate_args:
            self._validate_args()

    @classmethod
    def from_numbers(
        cls, loc: float | torch.Tensor, scale: float | torch.Tensor
    ) -> "NormalDistParams":
        return cls(*broadcast_all(loc, scale))

    def astuple(self) -> tuple[torch.Tensor, torch.Tensor]:
        return self.loc, self.scale

    @property
    def arg_constraints(self) -> dict[str, constraints.Constraint]:
        return self._arg_constraints

    @property
    def batch_shape(self) -> torch.Size:
        return self.loc.size()

    @property
    def event_shape(self) -> torch.Size:
        return torch.Size()

    @property
    def mean(self) -> torch.Tensor:
        return self.loc

    @property
    def mode(self) -> torch.Tensor:
        return self.loc

    @property
    def stddev(self) -> torch.Tensor:
        return self.scale

    @property
    def support(self) -> constraints.Constraint:
        return constraints.real

    @property
    def variance(self) -> torch.Tensor:
        return self.stddev.pow(2)


class Normal(Distribution[NormalDistParams]):
    @classmethod
    def sample(
        cls,
        params: NormalDistParams,
        sample_shape: torch.Size | tuple[int, ...] = torch.Size(),
        generator: torch.Generator | None = None,
    ) -> torch.Tensor:
        sample_shape = cls._extended_shape(params, sample_shape)
        with torch.no_grad():
            return torch.normal(
                params.loc.expand(sample_shape),
                params.scale.expand(sample_shape),
                generator=generator,
            )

    @classmethod
    def rsample(
        cls,
        params: NormalDistParams,
        sample_shape: torch.Size | tuple[int, ...] = torch.Size(),
        generator: torch.Generator | None = None,
    ) -> torch.Tensor:
        shape = cls._extended_shape(params, sample_shape)
        eps = _standard_normal(
            shape, dtype=params.loc.dtype, device=params.loc.device, generator=generator
        )
        return params.loc + eps * params.scale

    @classmethod
    def log_prob(
        cls,
        params: NormalDistParams,
        value: torch.Tensor,
    ) -> torch.Tensor:
        if params.validate_args:
            params.validate_sample(value)
        # compute the variance
        var = params.scale**2
        return -((value - params.loc) ** 2) / (2 * var) - 0.5 * torch.log(
            2 * var * math.pi
        )

    @classmethod
    def entropy(cls, params: NormalDistParams) -> torch.Tensor:
        return 0.5 + 0.5 * math.log(2 * math.pi) + torch.log(params.scale)



"""Type definitions / aliases for the sbft API."""
import typing as t

from pydantic import BaseModel

P = t.ParamSpec("P")
R = t.TypeVar("R")
R_co = t.TypeVar("R_co", covariant=True)
PydanticModel = t.TypeVar("PydanticModel", bound=BaseModel)


class EmptyModel(BaseModel):
    """A model with no fields."""

    class Config:
        allow_mutation = False


class BaseConfig(BaseModel):
    """Base config for all models."""

    class Config:
        allow_mutation = False
        arbitrary_types_allowed = True



"""Type definitions / aliases for the sbft API."""
import typing as t

import torch

from torchsilk import type_defs as tdf


class TorchModule(t.Protocol[tdf.P, tdf.R_co]):
    """A typed version of torch.nn.Module."""

    def forward(self, *args: tdf.P.args, **kwargs: tdf.P.kwargs) -> tdf.R_co:
        raise NotImplementedError


class FunctionalModuleWithoutBuffers(t.Protocol[tdf.P, tdf.R]):
    """A module that can be called like a function.

    A type-safe version of functorch's FunctionalModule which is obtained from the
    result of calling functorch.make_functional(...).
    """

    @classmethod
    def from_module(
        cls, module: TorchModule[tdf.P, tdf.R]
    ) -> "FunctionalModuleWithoutBuffers[tdf.P, tdf.R]":
        raise NotImplementedError

    def __call__(
        self,
        parameters: tuple[torch.nn.Parameter, ...],
        *args: tdf.P.args,
        **kwargs: tdf.P.kwargs,
    ) -> tdf.R:
        ...


class FunctionalModule(t.Protocol[tdf.P, tdf.R_co]):
    """A module that can be called like a function.

    A type-safe version of functorch's FunctionalModuleWithBuffers which is obtained
    from the result of calling functorch.make_functional_with_buffers(...).
    """

    @classmethod
    def from_module(
        cls, module: TorchModule[tdf.P, tdf.R]
    ) -> "FunctionalModuleWithoutBuffers[tdf.P, tdf.R]":
        raise NotImplementedError

    def __call__(
        self,
        parameters: tuple[torch.nn.Parameter, ...],
        buffers: tuple[torch.Tensor, ...],
        *args: tdf.P.args,
        **kwargs: tdf.P.kwargs,
    ) -> tdf.R_co:
        ...



__version__ = "0.5.0"

from pathlib import Path

REPO_ROOT = Path(__file__).parent.parent.expanduser().resolve()

from torchsilk import distributions as distributions  # noqa: E402
from torchsilk.modules import FunctionalModule as FunctionalModule  # noqa: E402
from torchsilk.modules import (  # noqa: E402
    FunctionalModuleWithoutBuffers as FunctionalModuleWithoutBuffers,
)
from torchsilk.modules import Module as Module  # noqa: E402
from torchsilk.modules import TrainState as TrainState  # noqa: E402
from torchsilk.type_defs import BaseConfig as BaseConfig  # noqa: E402
from torchsilk.type_defs import EmptyModel as EmptyModel  # noqa: E402



"""ModuleState is a container for a functional module along with its parameters and
buffers (if any)."""
import typing as t

import functorch as ft
import torch
import typing_extensions as te

from torchsilk import type_defs as tdf
from torchsilk.modules.buffers import NamedBuffers
from torchsilk.modules.type_defs import FunctionalModule, FunctionalModuleWithoutBuffers


class _ModuleStateWithoutBuffers(t.Generic[tdf.P, tdf.R_co]):
    """A container for a stateless module and its parameters.

    Example:

    .. code-block:: python

    >>> import torch
    ... import functorch as ft
    ... module = torch.nn.Sequential(
    ...     torch.nn.Linear(3, 3),
    ...     torch.nn.ReLU(),
    ...     torch.nn.Linear(3, 3),
    ... )
    ... x = torch.randn(4, 3)
    ... m = _ModuleStateWithoutBuffers.from_torch_module(module)
    ... exp_res = m(x)
    ... func_module, params = ft.make_functional(module)
    ... act_res = func_module(params, x)
    ... assert torch.allclose(exp_res, act_res)
    """

    def __init__(
        self,
        module: FunctionalModuleWithoutBuffers[tdf.P, tdf.R_co],
        parameters: tuple[torch.nn.Parameter, ...],
    ) -> None:
        """NOTE: Not intended to be called directly. Use `from_torch_module` instead."""
        self._module = module
        self._parameters = parameters

    @classmethod
    def from_torch_module(cls: type[te.Self], torch_module: torch.nn.Module) -> te.Self:
        """Create a `_ModuleStateWithoutBuffers` from a torch module.

        This is the main way to create a `_ModuleStateWithoutBuffers` instance.

        Example:

        .. code-block:: python

        >>> import torch
        ... import functorch as ft
        ... class MyModule(torch.nn.Module):
        ...     def __init__(self):
        ...         super().__init__()
        ...         self.linear = torch.nn.Linear(3, 3)
        ...         self.relu = torch.nn.ReLU()
        ...     def forward(self, x):
        ...         return self.relu(self.linear(x))
        ... module = MyModule()
        ... x = torch.randn(4, 3)
        ... m = _ModuleStateWithoutBuffers.from_torch_module(module)
        ... act_res = m(x)
        ... func_module, params = ft.make_functional(module)
        ... exp_res = func_module(params, x)
        ... assert torch.allclose(exp_res, act_res)

        >>> import torch
        ... import functorch as ft
        ... import pytest
        ... class MyModule(torch.nn.Module):
        ...     def __init__(self):
        ...         super().__init__()
        ...         self.linear = torch.nn.Linear(3, 2)
        ...         self.relu = torch.nn.ReLU()
        ...     def forward(self, x, y, *, z=0):
        ...         return self.relu(self.linear(x) + y) + z
        ... module = MyModule()
        ... x = torch.randn(4, 3)
        ... y = torch.randn(4, 2)
        ... z = torch.randn(4, 2)
        ... m = _ModuleStateWithoutBuffers.from_torch_module(module)
        ... act_res = m(x, y, z=z)
        ... func_module, params = ft.make_functional(module)
        ... exp_res = func_module(params, x, y, z=z)
        ... assert torch.allclose(exp_res, act_res)
        ... act_res = m(x, y)
        ... exp_res = func_module(params, x, y)
        ... assert torch.allclose(exp_res, act_res)
        ... also_act_res = m(x, y, z=1)
        ... also_exp_res = func_module(params, x, y, z=1)
        ... assert torch.allclose(also_exp_res, also_act_res)
        ... with pytest.raises(TypeError):
        ...     m(x, y, z=z, w=1)
        """
        func, parameters = ft.make_functional(
            torch_module, disable_autograd_tracking=True
        )
        return cls(
            t.cast(FunctionalModuleWithoutBuffers[tdf.P, tdf.R_co], func), parameters
        )

    @property
    def module(self) -> FunctionalModuleWithoutBuffers[tdf.P, tdf.R_co]:
        """Return the underlying stateless module."""
        return self._module

    @property
    def parameters(self) -> tuple[torch.nn.Parameter, ...]:
        """Return the parameters of the underlying stateless module."""
        return self._parameters

    def apply(
        self,
        params: tuple[torch.nn.Parameter, ...],
        *args: tdf.P.args,
        **kwargs: tdf.P.kwargs,
    ) -> tdf.R_co:
        """Apply the module to the given parameters and buffers.

        Example:

        >>> import torch
        ... import functorch as ft
        ... module = torch.nn.Sequential(
        ...     torch.nn.Linear(3, 3),
        ...     torch.nn.ReLU(),
        ...     torch.nn.Linear(3, 3),
        ... )
        ... x = torch.randn(4, 3)
        ... m = _ModuleStateWithoutBuffers.from_torch_module(module)
        ... exp_res = m(x)
        ... func_module, params = ft.make_functional(module)
        ... act_res = func_module(params, x)
        ... assert torch.allclose(exp_res, act_res)
        ... new_params = tuple(p + 1 for p in params)
        ... new_act_res = m.apply(new_params, x)
        ... new_exp_res = func_module(new_params, x)
        ... assert torch.allclose(new_exp_res, new_act_res)
        """
        return self._module(params, *args, **kwargs)

    def evolve(
        self,
        parameters: tuple[torch.nn.Parameter, ...],
    ) -> te.Self:
        """Return a new `_ModuleStateWithoutBuffers` with the given changes.

        Example:

        >>> import torch
        ... import functorch as ft
        ... class MyModule(torch.nn.Module):
        ...     def __init__(self):
        ...         super().__init__()
        ...         self.linear = torch.nn.Linear(3, 3)
        ...         self.relu = torch.nn.ReLU()
        ...     def forward(self, x):
        ...         return self.relu(self.linear(x))
        ... module = MyModule()
        ... x = torch.randn(4, 3)
        ... m = _ModuleStateWithoutBuffers.from_torch_module(module)
        ... func_module, params = ft.make_functional(module)
        ... exp_res = func_module(params, x)
        ... act_res = m(x)
        ... assert torch.allclose(exp_res, act_res)
        ... new_params = tuple(p + 1 for p in params)
        ... m = m.evolve(new_params)
        ... new_res = m(x)
        ... assert not torch.allclose(exp_res, new_res)
        """
        if parameters is None:
            parameters = self._parameters
        return self.__class__(self.module, parameters)

    def __call__(self, *args: tdf.P.args, **kwargs: tdf.P.kwargs) -> tdf.R_co:
        return self.module(self.parameters, *args, **kwargs)


class _ModuleStateWithBuffers(t.Generic[tdf.P, tdf.R_co]):
    """A container for a stateless module, its parameters and its buffers.

    Example:

    .. code-block:: python

    >>> import torch
    ... import functorch as ft
    ... class ModuleWithBuffers(torch.nn.Module):
    ...     def __init__(self):
    ...         super().__init__()
    ...         self.register_buffer("x", torch.randn(3))
    ...     def forward(self, x):
    ...         return x + self.x
    ... module = ModuleWithBuffers()
    ... x = torch.randn(4, 3)
    ... m = _ModuleStateWithBuffers.from_torch_module(module)
    ... act_res = m(x)
    ... func_module, params, buffers = ft.make_functional_with_buffers(module)
    ... exp_res = func_module(params, buffers, x)
    ... assert torch.allclose(exp_res, act_res)
    ... assert torch.allclose(m.named_buffers.x, buffers[0])
    """

    def __init__(
        self,
        module: FunctionalModule[tdf.P, tdf.R_co],
        parameters: tuple[torch.nn.Parameter, ...],
        buffers: tuple[torch.Tensor, ...],
        named_buffers: NamedBuffers,
    ) -> None:
        self._module = module
        self._parameters = parameters
        self._buffers = buffers
        self._named_buffers = named_buffers

    @classmethod
    def from_torch_module(cls: type[te.Self], torch_module: torch.nn.Module) -> te.Self:
        """Create a `_ModuleStateWithBuffers` from a torch module.

        This is the main way to create a `_ModuleStateWithBuffers` instance.

        Examples:

        .. code-block:: python

        >>> import torch
        ... import functorch as ft
        ... class ModuleWithBuffers(torch.nn.Module):
        ...     def __init__(self):
        ...         super().__init__()
        ...         self.register_buffer("x", torch.randn(3))
        ...     def forward(self, x):
        ...         return x + self.x
        ... module = ModuleWithBuffers()
        ... x = torch.randn(4, 3)
        ... m = _ModuleStateWithBuffers.from_torch_module(module)
        ... act_res = m(x)
        ... func_module, params, buffers = ft.make_functional_with_buffers(module)
        ... exp_res = func_module(params, buffers, x)
        ... assert torch.allclose(exp_res, act_res)
        ... assert torch.allclose(m.named_buffers.x, buffers[0])
        """
        named_buffers = NamedBuffers.from_module(torch_module)
        func, parameters, buffers = ft.make_functional_with_buffers(
            torch_module, disable_autograd_tracking=True
        )
        return cls(
            module=t.cast(FunctionalModule[tdf.P, tdf.R_co], func),
            parameters=parameters,
            buffers=buffers,
            named_buffers=named_buffers.evolve_from_values(buffers),
        )

    @property
    def module(self) -> FunctionalModule[tdf.P, tdf.R_co]:
        return self._module

    @property
    def parameters(self) -> tuple[torch.nn.Parameter, ...]:
        return self._parameters

    @property
    def buffers(self) -> tuple[torch.Tensor, ...]:
        return self._buffers

    @property
    def named_buffers(self) -> NamedBuffers:
        return self._named_buffers

    def apply(
        self,
        params: tuple[torch.nn.Parameter, ...],
        buffers: tuple[torch.Tensor, ...],
        *args: tdf.P.args,
        **kwargs: tdf.P.kwargs,
    ) -> tdf.R_co:
        """Apply the module to the given parameters and buffers.

        Example:

        >>> import torch
        ... import functorch as ft
        ... class ModuleWithBuffers(torch.nn.Module):
        ...     def __init__(self):
        ...         super().__init__()
        ...         self.linear = torch.nn.Linear(3, 3)
        ...         self.register_buffer("x", torch.randn(3))
        ...     def forward(self, x):
        ...         return self.linear(x) + self.x
        ... module = ModuleWithBuffers()
        ... x = torch.randn(4, 3)
        ... m = _ModuleStateWithBuffers.from_torch_module(module)
        ... exp_res = m(x)
        ... func_module, params, buffers = ft.make_functional_with_buffers(module)
        ... act_res = func_module(params, buffers, x)
        ... assert torch.allclose(exp_res, act_res)
        ... new_params = tuple(p + 1 for p in params)
        ... new_act_res = m.apply(new_params, buffers, x)
        ... new_exp_res = func_module(new_params, buffers, x)
        ... assert torch.allclose(new_exp_res, new_act_res)
        """
        return self._module(params, buffers, *args, **kwargs)

    def evolve(
        self: te.Self,
        parameters: tuple[torch.nn.Parameter, ...] | None = None,
        buffers: tuple[torch.Tensor, ...] | None = None,
        named_buffers: NamedBuffers | None = None,
    ) -> te.Self:
        """Return a new `_ModuleStateWithBuffers` with the given changes.

        Args:
        -----
            parameters: tuple[torch.nn.Parameter, ...] | None
                The new parameters of the module. If None, parameters are not changed.
            buffers: tuple[torch.Tensor, ...] | None
                The new buffers of the module. If None, buffers are not changed.
                Cannot be used together with named_buffers.
            named_buffers: NamedBuffers | None
                The new named buffers of the module. If None, named_buffers are not
                changed.
                Cannot be used together with buffers.
        Returns:
        --------
            A new `_ModuleStateWithBuffers` with the given changes.

        Examples:

        >>> import torch
        ... import functorch as ft
        ... class ModuleWithBuffers(torch.nn.Module):
        ...     def __init__(self):
        ...         super().__init__()
        ...         self.register_buffer("x", torch.randn(3))
        ...     def forward(self, x):
        ...         return x + self.x
        ... module = ModuleWithBuffers()
        ... x = torch.randn(4, 3)
        ... m = _ModuleStateWithBuffers.from_torch_module(module)
        ... act_res = m(x)
        ... func_module, params, buffers = ft.make_functional_with_buffers(module)
        ... exp_res = func_module(params, buffers, x)
        ... assert torch.allclose(exp_res, act_res)
        ... assert torch.allclose(m.named_buffers.x, buffers[0])
        ... m = m.evolve(buffers=(torch.randn(3),))
        ... new_res = m(x)
        ... assert not torch.allclose(new_res, act_res)

        # Cannot use both buffers and named_buffers
        >>> m = m.evolve(
        ...     buffers=(torch.randn(3),), named_buffers=m.named_buffers
        ... )  # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
        ValueError: Cannot specify both buffers and named_buffers
        """
        if parameters is None:
            parameters = self.parameters
        if buffers and named_buffers:
            raise ValueError("Cannot specify both buffers and named_buffers")
        elif buffers:
            named_buffers = self.named_buffers.evolve_from_values(buffers)
        elif named_buffers:
            buffers = tuple(named_buffers.values())
        else:
            buffers = self.buffers
            named_buffers = self.named_buffers
        assert parameters is not None
        assert buffers is not None
        assert named_buffers is not None
        return self.__class__(
            module=self._module,
            parameters=parameters,
            buffers=buffers,
            named_buffers=named_buffers,
        )

    def __call__(self, *args: tdf.P.args, **kwargs: tdf.P.kwargs) -> tdf.R_co:
        return self.module(self.parameters, self.buffers, *args, **kwargs)


class ModuleState(t.Generic[tdf.P, tdf.R_co]):
    """A container for a stateless module, its parameters and its buffers if it has any.

    NOTE: Do not initialize this directly. Use ModuleState.from_torch_module.

    Example:
    1. A module without buffers
    >>> import torch
    ... import functorch as ft
    ... class ModuleWithoutBuffers(torch.nn.Module):
    ...     def __init__(self):
    ...         super().__init__()
    ...         self.linear = torch.nn.Linear(3, 4)
    ...     def forward(self, x, y):
    ...         return self.linear(x) + y
    ... torch_module = ModuleWithoutBuffers()
    ... module_state = ModuleState.from_torch_module(torch_module)
    ... x = torch.randn(4, 3)
    ... y = torch.randn(4, 4)
    ... act_res = module_state(x, y)
    ... func_module, params = ft.make_functional(torch_module)
    ... exp_res = func_module(params, x, y)
    ... assert torch.allclose(exp_res, act_res)

    2. A module with buffers
    >>> import torch
    ... import functorch as ft
    ... class ModuleWithBuffers(torch.nn.Module):
    ...     def __init__(self):
    ...         super().__init__()
    ...         self.register_buffer("x", torch.randn(3))
    ...     def forward(self, x):
    ...         return x + self.x
    ... torch_module = ModuleWithBuffers()
    ... module_state = ModuleState.from_torch_module(torch_module)
    ... x = torch.randn(4, 3)
    ... act_res = module_state(x)
    ... func_module, params, buffers = ft.make_functional_with_buffers(torch_module)
    ... exp_res = func_module(params, buffers, x)
    ... assert torch.allclose(exp_res, act_res)
    ... assert torch.allclose(module_state.named_buffers.x, buffers[0])
    """

    def __init__(
        self,
        module_state: _ModuleStateWithoutBuffers[tdf.P, tdf.R_co]
        | _ModuleStateWithBuffers[tdf.P, tdf.R_co],
    ) -> None:
        self._module_state = module_state
        self._has_buffers = isinstance(module_state, _ModuleStateWithBuffers)
        self._buffers = (
            module_state.buffers
            if isinstance(module_state, _ModuleStateWithBuffers)
            else t.cast(tuple[torch.Tensor, ...], tuple())
        )
        self._named_buffers = (
            module_state.named_buffers
            if isinstance(module_state, _ModuleStateWithBuffers)
            else NamedBuffers({})
        )

    @classmethod
    def from_torch_module(cls: type[te.Self], torch_module: torch.nn.Module) -> te.Self:
        if next(torch_module.buffers(recurse=True), None) is not None:
            # If the module has buffers, use _ModuleStateWithBuffers.
            return cls(
                _ModuleStateWithBuffers[tdf.P, tdf.R_co].from_torch_module(torch_module)
            )
        else:
            # Otherwise, use _ModuleStateWithoutBuffers.
            return cls(
                _ModuleStateWithoutBuffers[tdf.P, tdf.R_co].from_torch_module(
                    torch_module
                )
            )

    @property
    def module(
        self,
    ) -> FunctionalModuleWithoutBuffers[tdf.P, tdf.R_co] | FunctionalModule[
        tdf.P, tdf.R_co
    ]:
        return self._module_state.module

    @property
    def parameters(self) -> tuple[torch.nn.Parameter, ...]:
        return self._module_state.parameters

    @property
    def buffers(self) -> tuple[torch.Tensor, ...]:
        return self._buffers

    @property
    def named_buffers(self) -> NamedBuffers:
        return self._named_buffers

    @property
    def has_buffers(self) -> bool:
        return self._has_buffers

    def apply(
        self,
        params: tuple[torch.nn.Parameter, ...],
        *args: tdf.P.args,
        **kwargs: tdf.P.kwargs,
    ) -> tdf.R_co:
        """Applies the module with the given parameters and buffers.

        NOTE: This is equivalent to calling the module state directly.

        Example:
        1. A module without buffers
        >>> import torch
        ... import functorch as ft
        ... class ModuleWithoutBuffers(torch.nn.Module):
        ...     def __init__(self):
        ...         super().__init__()
        ...         self.linear = torch.nn.Linear(3, 4)
        ...     def forward(self, x, y):
        ...         return self.linear(x) + y
        ... torch_module = ModuleWithoutBuffers()
        ... module_state = ModuleState.from_torch_module(torch_module)
        ... x = torch.randn(4, 3)
        ... y = torch.randn(4, 4)
        ... act_res = module_state.apply(module_state.parameters, x, y)
        ... func_module, params = ft.make_functional(torch_module)
        ... exp_res = func_module(params, x, y)
        ... assert torch.allclose(exp_res, act_res)

        2. A module with buffers
        >>> import torch
        ... import functorch as ft
        ... class ModuleWithBuffers(torch.nn.Module):
        ...     def __init__(self):
        ...         super().__init__()
        ...         self.linear = torch.nn.Linear(3, 3)
        ...         self.register_buffer("x", torch.randn(3))
        ...     def forward(self, x):
        ...         return self.linear(x) + self.x
        ... torch_module = ModuleWithBuffers()
        ... module_state = ModuleState.from_torch_module(torch_module)
        ... func_module, *_ = ft.make_functional_with_buffers(torch_module)
        ... x = torch.randn(4, 3)
        ... act_res = module_state.apply(
        ...     module_state.parameters, x
        ... )
        ... exp_res = func_module(module_state.parameters, module_state.buffers, x)
        ... assert torch.allclose(exp_res, act_res)
        ... new_params = tuple(p + 1 for p in module_state.parameters)
        ... new_act_res = module_state.apply(new_params, x)
        ... assert not torch.allclose(new_act_res, act_res)
        ... new_exp_res = func_module(new_params, module_state.buffers, x)
        ... assert torch.allclose(new_exp_res, new_act_res)
        """
        if self.has_buffers:
            assert isinstance(self._module_state, _ModuleStateWithBuffers)
            return self._module_state.apply(params, self.buffers, *args, **kwargs)
        assert isinstance(self._module_state, _ModuleStateWithoutBuffers)
        return self._module_state.apply(params, *args, **kwargs)

    def apply_with_buffers(
        self,
        params: tuple[torch.nn.Parameter, ...],
        buffers: tuple[torch.Tensor, ...],
        *args: tdf.P.args,
        **kwargs: tdf.P.kwargs,
    ) -> tdf.R_co:
        """Applies the module with the given parameters and buffers.

        NOTE: This is equivalent to calling the module state directly.

        Example:
        1. A module without buffers
        >>> import torch
        ... import functorch as ft
        ... import pytest
        ... class ModuleWithoutBuffers(torch.nn.Module):
        ...     def __init__(self):
        ...         super().__init__()
        ...         self.linear = torch.nn.Linear(3, 4)
        ...     def forward(self, x, y):
        ...         return self.linear(x) + y
        ... torch_module = ModuleWithoutBuffers()
        ... module_state = ModuleState.from_torch_module(torch_module)
        ... func_module, *_ = ft.make_functional(torch_module)
        ... x = torch.randn(4, 3)
        ... y = torch.randn(4, 4)
        ... act_res = module_state.apply(module_state.parameters, x, y)
        ... exp_res = func_module(module_state.parameters, x, y)
        ... assert torch.allclose(exp_res, act_res)

        2. A module with buffers
        >>> import torch
        ... import functorch as ft
        ... class ModuleWithBuffers(torch.nn.Module):
        ...     def __init__(self):
        ...         super().__init__()
        ...         self.linear = torch.nn.Linear(3, 3)
        ...         self.register_buffer("x", torch.randn(3))
        ...     def forward(self, x):
        ...         return self.linear(x) + self.x
        ... torch_module = ModuleWithBuffers()
        ... module_state = ModuleState.from_torch_module(torch_module)
        ... func_module, *_ = ft.make_functional_with_buffers(torch_module)
        ... x = torch.randn(4, 3)
        ... act_res = module_state.apply_with_buffers(
        ...     module_state.parameters, module_state.buffers, x
        ... )
        ... exp_res = func_module(module_state.parameters, module_state.buffers, x)
        ... assert torch.allclose(exp_res, act_res)
        ... new_params = tuple(p + 1 for p in module_state.parameters)
        ... new_buffers = tuple(b + 1 for b in module_state.buffers)
        ... new_act_res = module_state.apply_with_buffers(new_params, new_buffers, x)
        ... assert not torch.allclose(new_act_res, act_res)
        ... new_exp_res = func_module(new_params, new_buffers, x)
        ... assert torch.allclose(new_exp_res, new_act_res)
        """
        if not self.has_buffers:
            raise ValueError("This module state does not have buffers.")
        assert isinstance(self._module_state, _ModuleStateWithBuffers)
        return self._module_state.apply(params, buffers, *args, **kwargs)

    def evolve(
        self,
        parameters: tuple[torch.nn.Parameter, ...] | None = None,
        buffers: tuple[torch.Tensor, ...] | None = None,
        named_buffers: NamedBuffers | None = None,
    ) -> "ModuleState[tdf.P, tdf.R_co]":
        """Return a new ModuleState with the specified changes.

        Args:
        -----
            parameters: tuple[torch.nn.Parameter, ...] | None
                If not None, update the parameters.
            buffers: tuple[torch.Tensor, ...] | None
                If not None, update the buffers.
            named_buffers: NamedBuffers | None
                If not None, update the named_buffers.
        Returns:
        --------
            A new ModuleState with the specified changes.
        Raises:
        -------
            ValueError: If both buffers and named_buffers are specified.
            ValueError: If module has no buffers and no parameters are specified.

        Example:
        1. Update parameters
        >>> import torch
        ... import functorch as ft
        ... class ModuleWithoutBuffers(torch.nn.Module):
        ...     def __init__(self):
        ...         super().__init__()
        ...         self.linear = torch.nn.Linear(3, 4)
        ...     def forward(self, x, y):
        ...         return self.linear(x) + y
        ... torch_module = ModuleWithoutBuffers()
        ... module_state = ModuleState.from_torch_module(torch_module)
        ... x = torch.randn(4, 3)
        ... y = torch.randn(4, 4)
        ... act_res = module_state(x, y)
        ... func_module, params = ft.make_functional(torch_module)
        ... exp_res = func_module(params, x, y)
        ... assert torch.allclose(exp_res, act_res)
        ... new_params = tuple(p + 1 for p in params)
        ... new_module_state = module_state.evolve(parameters=new_params)
        ... new_res = new_module_state(x, y)
        ... assert not torch.allclose(exp_res, new_res)

        2. Update buffers
        >>> import torch
        ... import functorch as ft
        ... class ModuleWithBuffers(torch.nn.Module):
        ...     def __init__(self):
        ...         super().__init__()
        ...         self.register_buffer("x", torch.randn(3))
        ...     def forward(self, x):
        ...         return x + self.x
        ... torch_module = ModuleWithBuffers()
        ... module_state = ModuleState.from_torch_module(torch_module)
        ... x = torch.randn(4, 3)
        ... act_res = module_state(x)
        ... func_module, params, buffers = ft.make_functional_with_buffers(torch_module)
        ... exp_res = func_module(params, buffers, x)
        ... assert torch.allclose(exp_res, act_res)
        ... new_buffers = tuple(b + 1 for b in buffers)
        ... new_module_state = module_state.evolve(buffers=new_buffers)
        ... new_res = new_module_state(x)
        ... assert not torch.allclose(exp_res, new_res)

        3. Passing both buffers and named_buffers raises ValueError
        >>> import torch
        ... import functorch as ft
        ... import pytest
        ... class ModuleWithBuffers(torch.nn.Module):
        ...     def __init__(self):
        ...         super().__init__()
        ...         self.register_buffer("x", torch.randn(3))
        ...     def forward(self, x):
        ...         return x + self.x
        ... torch_module = ModuleWithBuffers()
        ... module_state = ModuleState.from_torch_module(torch_module)
        ... x = torch.randn(4, 3)
        ... act_res = module_state(x)
        ... func_module, params, buffers = ft.make_functional_with_buffers(torch_module)
        ... exp_res = func_module(params, buffers, x)
        ... assert torch.allclose(exp_res, act_res)
        ... new_buffers = tuple(b + 1 for b in buffers)
        ... new_named_buffers = module_state.named_buffers.evolve_from_values(
        ...     new_buffers
        ... )
        ... with pytest.raises(ValueError):
        ...     module_state.evolve(
        ...         buffers=new_buffers, named_buffers=new_named_buffers
        ...     )
        """
        if not self.has_buffers:
            if not parameters:
                raise ValueError(
                    "parameters must be specified when the module has no buffers."
                )
            return self.__class__(self._module_state.evolve(parameters=parameters))
        assert isinstance(self._module_state, _ModuleStateWithBuffers)
        if buffers and named_buffers:
            raise ValueError("Cannot specify both buffers and named_buffers")
        return self.__class__(
            self._module_state.evolve(
                parameters=parameters, buffers=buffers, named_buffers=named_buffers
            )
        )

    def to(self, *args: t.Any, **kwargs: t.Any) -> "ModuleState[tdf.P, tdf.R_co]":
        """Move the buffers to the specified device.

        Args:
        -----
            *args: Any
            **kwargs: Any
            Any arguments accepted by torch.Tensor.to(...).
        Returns:
        --------
            A new ModuleState with the buffers moved to the specified device and/or
            dtype and/or memory format.

        Examples:
        ---------

        >>> import torch
        ... import functorch as ft
        ... import torchsilk as tsk
        ... class ModuleWithBuffers(torch.nn.Module):
        ...     def __init__(self):
        ...         super().__init__()
        ...         self.register_buffer("x", torch.randn(3))
        ...     def forward(self, x):
        ...         return x + self.x
        ... torch_module = ModuleWithBuffers()
        ... module_state = ModuleState.from_torch_module(torch_module)
        ... float64_module_state = module_state.to(dtype=torch.float64)
        ... assert float64_module_state.named_buffers.x.dtype == torch.float64
        """
        new_params = t.cast(
            tuple[torch.nn.Parameter, ...],
            tuple(x.to(*args, **kwargs) for x in self.parameters),
        )
        new_named_buffers = {
            k: v.to(*args, **kwargs) for k, v in self.named_buffers.items()
        }
        new_buffers = tuple(new_named_buffers.values())
        return self.__class__(
            module_state=_ModuleStateWithBuffers(
                module=self.module,  # type: ignore
                parameters=new_params,
                buffers=new_buffers,
                named_buffers=NamedBuffers(new_named_buffers),
            )
            if new_buffers
            else _ModuleStateWithoutBuffers(
                module=self.module, parameters=new_params  # type: ignore
            )
        )

    def __call__(self, *args: tdf.P.args, **kwargs: tdf.P.kwargs) -> tdf.R_co:
        return self._module_state(*args, **kwargs)

    def __str__(self) -> str:
        return str(self._module_state)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['torchsilk', 'torchsilk.distributions', 'torchsilk.modules']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=22.1.0,<23.0.0',
 'lovely-tensors>=0.1.8,<0.2.0',
 'numpy>=1.23.5,<2.0.0',
 'optree>=0.5.0,<0.6.0',
 'pydantic>=1.10.2,<2.0.0',
 'pyyaml>=6.0,<7.0',
 'rich>=12.6.0,<13.0.0',
 'scipy>=1.9.3,<2.0.0',
 'torch>=1.13.0,<2.0.0',
 'torchopt>=0.5.0.post4,<0.6.0',
 'typer[all]>=0.7.0,<0.8.0']

extras_require = \
{'all': ['autoflake>=1.4,<2.0',
         'black>=22.3.0,<23.0.0',
         'isort>=5.10.1,<6.0.0',
         'mypy>=0.961,<0.962',
         'pytest>=7.2.0,<8.0.0',
         'pytest-vscodedebug>=0.1.0,<0.2.0',
         'pyupgrade>=2.37.3,<3.0.0',
         'toml>=0.10.2,<0.11.0',
         'xdoctest[all]>=1.1.0,<2.0.0'],
 'test': ['autoflake>=1.4,<2.0',
          'black>=22.3.0,<23.0.0',
          'isort>=5.10.1,<6.0.0',
          'mypy>=0.961,<0.962',
          'pytest>=7.2.0,<8.0.0',
          'pytest-vscodedebug>=0.1.0,<0.2.0',
          'pyupgrade>=2.37.3,<3.0.0',
          'toml>=0.10.2,<0.11.0',
          'xdoctest[all]>=1.1.0,<2.0.0']}

setup_kwargs = {
    'name': 'torchsilk',
    'version': '0.5.0',
    'description': 'TorchSilk is a neural network library for working with functorch.',
    'long_description': '# TorchSilk\n\nTorchSilk is a neural network library for working with functorch.\n',
    'author': 'Aditya Gudimella',
    'author_email': 'aditya.gudimella@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.10,<3.11',
}


setup(**setup_kwargs)
